"""Defines the logic for the core game play."""
import math
import random
import os.path

from cocos.sprite import Sprite
import pyglet

import soundex
from constants import *

# =============================================================================

def get_vector(v):
    return (v[0], v[1])

def make_vector(x, y):
    return (x, y)

def add_vectors(v1, v2):
    return (v1[0] + v2[0], v1[1] + v2[1])

def add_vector_and_direction(v, direction, length):
    if direction == 0:
        return (v[0], v[1] - length)
    elif direction == 1:
        return (v[0] + length, v[1])
    elif direction == 2:
        return (v[0], v[1] + length)
    else:
        return (v[0] - length, v[1])

def transform_grid_to_screen_position(v):
    return make_vector(
        GRID_X_OFFSET + 36/2 + 36 * v[0],
        GRID_HEIGHT + GRID_Y_OFFSET - 36/2 - 36 * v[1])


# =============================================================================
#
# ActorEvent and actor events
#
# =============================================================================

class ActorEvent(object):
    """An action or happening, which can occur and be applied to the certain actor.

    The history of actor events are remembered in the timeline-like fashion.
    Events can be both undone while time travel and replayed.
    """
    def __init__(self, time, duration):
        self.begin_time = time
        self.end_time = time + duration

class MovementEvent(ActorEvent):
    """An action of actor movement."""
    def __init__(self, time, duration, direction):
        super(MovementEvent, self).__init__(time, duration)
        self.direction = direction

class SwitchEvent(ActorEvent):
    """An action of gate switching."""
    def __init__(self, time, switch_state):
        super(SwitchEvent, self).__init__(time, 0)
        self.switch_state = switch_state

# =============================================================================
#
# Actor and actors
#
# =============================================================================

class Actor(object):
    """A game object.

    Actors can be affected by action and events, which are contained a timeline.
    """
    def __init__(self, grid_position):
        self.level = None
        self.creation_time = 0
        self.lived_time = -1
        self.grid_position = get_vector(grid_position)
        self.exact_position = get_vector(grid_position)
        self.is_moving = False # == not grid-aligned.
        self.event_timeline = []
        self.current_event = None
    def render(self):
        pass
    def get_event_by_time_for_replay(self, time):
        for event in self.event_timeline:
            if time >= event.begin_time and time < event.end_time:
                return event
        return None
    def get_event_by_time_for_time_travel(self, time, remove):
        if self.event_timeline:
            last_index = len(self.event_timeline)
        while self.event_timeline:
            last_index -= 1
            if last_index < 0:
                break
            event = self.event_timeline[last_index]
            if time > event.end_time:
                break
            elif time > event.begin_time:
                return event
            elif remove:
                del self.event_timeline[last_index]
        return None
    def update(self):
        if self.level.play_time > self.lived_time:
            self.lived_time = self.level.play_time
        if self.is_moving:
            self.exact_position = add_vector_and_direction(self.exact_position, self.current_event.direction, 0.1)
            if self.level.play_time >= self.current_event.end_time:
                self.is_moving = False
                self.current_event = None
                self.exact_position = get_vector(self.grid_position)
            self.s.set_position(*transform_grid_to_screen_position(self.exact_position))

    def update_time_travel(self):
        # Replay movements in the backward direction.
        event = self.get_event_by_time_for_time_travel(self.level.play_time, not isinstance(self, Avatar))
        self.current_event = event
        if event:
            self.is_moving = True
            if isinstance(event, MovementEvent):
                if self.level.play_time == event.end_time:
                    self.grid_position = add_vector_and_direction(self.grid_position, (event.direction + 2) % 4, 1)
                self.exact_position = add_vector_and_direction(self.exact_position, (event.direction + 2) % 4, 0.1)
                self.s.set_position(*transform_grid_to_screen_position(self.exact_position))
        else:
            self.is_moving = False

    def make_move(self, direction, force=False):
        """Command the actor to move in the specified direction in the specified time."""
        if self.is_moving:
            return ("moving", None)

        avatar_replay = False
        if self.current_event or isinstance(self, Avatar) and self.level.play_time <= self.lived_time:
            avatar_replay = True

        destination_position = add_vector_and_direction(self.grid_position, direction, 1)
        destination_tile = self.level.tile_grid[destination_position[1]][destination_position[0]]
        colliders = self.level.gather_actors_in_tile(destination_position)

        # Analyze the colliding actors and determine whether the move can be
        # accomplished.
        can_proceed = False
        if force:
            can_proceed = True
        elif destination_tile == ' ':
            if not colliders:
                can_proceed = True
            else:
                can_proceed = True
                for collider in colliders:
                    if not collider.is_moving:
                        can_proceed = False
                    elif not isinstance(collider.current_event, MovementEvent):
                        can_proceed = False
                    elif collider.current_event.direction != direction:
                        can_proceed = False
                    elif collider.current_event.begin_time == self.level.play_time:
                        can_proceed = False

        if can_proceed:
            if avatar_replay:
                self.current_event = self.get_event_by_time_for_replay(self.level.play_time)
            else:
                self.current_event = MovementEvent(self.level.play_time, 10, direction)
                self.event_timeline.append(self.current_event)
            self.is_moving = True
            self.grid_position = destination_position
            return ("succeed", None)
        else:
            return ("blocked", colliders)

class Boulder(Actor):
    """A boulder object.

    Boulders can be pushed. They stay still until touched. After being touched
    they fall downwards.
    """
    def __init__(self, position):
        super(Boulder, self).__init__(position)
        self.is_falling = False
        self.touch_time = -1
    def render(self):
        self.s = Sprite('boulder1a.png', position=transform_grid_to_screen_position(self.grid_position))
        self.level.layer.add(self.s)
    def update(self):
        super(Boulder, self).update()
        if self.is_falling:
            self.make_move(2)
    def update_time_travel(self):
        super(Boulder, self).update_time_travel()
        if self.level.play_time <= self.touch_time:
            self.is_falling = False
            self.touch_time = -1

class Crate(Actor):
    """A crate object.

    Crate can be pushed. Unlike boulders, they do not fall.
    """
    def __init__(self, position):
        super(Crate, self).__init__(position)
    def render(self):
        self.s = Sprite('crate1a.png', position=transform_grid_to_screen_position(self.grid_position))
        self.level.layer.add(self.s)

class Avatar(Actor):
    """The player's main character.

    The Professor When himself!
    """
    def __init__(self, position):
        super(Avatar, self).__init__(position)
    def render(self):
        # TODO: Render the avatar at self.exact_position. self.level.play_time designates the animation frame.
        self.s = Sprite('prof1a.png', position=transform_grid_to_screen_position(self.grid_position))
        self.level.layer.add(self.s)
    def update(self):
        if self.level.play_time <= self.lived_time:
            self.current_event = self.get_event_by_time_for_replay(self.level.play_time)
            if not self.current_event:
                self.is_moving = False
            if self.current_event and self.level.play_time == self.current_event.begin_time:
                move_result = self.move_avatar(self.current_event.direction)
                if not move_result:
                    self.level.continuum_broke = True
                    print "Continuum has been broked."

        super(Avatar, self).update()

        # Avatar below the boulder makes it start falling.
        if not self.is_moving:
            for actor in self.level.gather_actors_in_tile(add_vector_and_direction(self.grid_position, 0, 1)):
                if isinstance(actor, Boulder):
                    actor.is_falling = True
                    actor.touch_time = self.level.play_time

    def move_avatar(self, direction):
        """Try to begin the avatar movement in the specified direction.

        direction       0 - Up.
                        1 - Right.
                        2 - Down.
                        3 - Left.
        """
        if self.level.travelling_time:
            print "Restoring the time flow."
        self.level.travelling_time = False

        (move_result, colliders) = self.make_move(direction)
        if move_result == "succeed":
            return True

        if move_result == "blocked" and len(colliders) == 1:
            if isinstance(colliders[0], TimeWarp):
                self.make_move(direction, force=True)
                return True
            elif isinstance(colliders[0], LevelExit):
                self.level.layer.flash()
                self.make_move(direction, force=True)
                return True
            elif isinstance(colliders[0], Crate):
                crate = colliders[0]
                (crate_move_result, crate_colliders) = crate.make_move(direction)
                if crate_move_result == "succeed":
                    soundex.play("wood_move.wav")
                    self.make_move(direction, force=True)
                    return True
            elif isinstance(colliders[0], Boulder):
                boulder = colliders[0]
                boulder.is_falling = True
                boulder.touch_time = self.level.play_time
                if direction == 1 or direction == 2 or direction == 3:
                    (boulder_move_result, boulder_colliders) = boulder.make_move(direction)
                    if boulder_move_result == "succeed":
                        soundex.play("rock_move.wav")
                        self.make_move(direction, force=True)
                        return True
        return False

class LevelExit(Actor):
    """The blue police phone booth object."""
    def __init__(self, position):
        super(LevelExit, self).__init__(position)
    def render(self):
        images = (pyglet.image.load(os.path.join("data", "booth1%s.png" % c)) for c in "abcb")
        animation = pyglet.image.Animation.from_image_sequence(images, 0.1)
        s = Sprite(animation, position=transform_grid_to_screen_position(self.grid_position))
        self.level.layer.add(s)

class TimeWarp(Actor):
    """A time warp object."""
    def __init__(self, position):
        super(TimeWarp, self).__init__(position)
    def render(self):
        images = (pyglet.image.load(os.path.join("data", "energy1%s1.png" % c)) for c in "abcd")
        animation = pyglet.image.Animation.from_image_sequence(images, 0.1)
        s = Sprite(animation, position=transform_grid_to_screen_position(self.grid_position))
        self.level.layer.add(s)

class Gate(Actor):
    """A gate object."""
    def __init__(self, position):
        super(Gate, self).__init__(position)
    def render(self):
        s = Sprite('gate1a.png', position=transform_grid_to_screen_position(self.grid_position))
        self.level.layer.add(s)

# =============================================================================
#
# LevelEngine
#
# =============================================================================

class LevelEngine(object):
    """Base class for all the levels defined in the game.

    Implements the game logic.
    """

    def __init__(self):
        self.map_name = None
        self.background_image = None
        self.tile_grid = None
        self.actors = None
        self.play_time = 0 # Time is discreet and is represented by the frame count.
        self.travelling_time = False
        self.avatar = None
        self.play_state = "start"
        self.continuum_broke = False

    def set_map_plan(self, map_plan):
        """Read the map plan string and generate the tile grid and the actor collection.

        The results are stored in self.tile_grid and self.actor_models.

        map_plan        A string representing the map plan. It's lines
                        designate rows of the map. Each character represents
                        the game object, like wall tile.
        """
        # @type map_plan str

        tile_grid = []
        max_tile_grid_row_length = 0
        actors = []

        # Scan the map plan by lines.
        tile_location = [0, 0] # [x, y]
        for plan_line in map_plan.splitlines():
            # @type map_line str
            if not plan_line.strip():
                # The line contains only white characters. Ignore such ones.
                continue
            tile_location[0] = 0
            tile_grid_row = []
            for plan_char in plan_line:
                if plan_char == ' ':
                    # Empty space tile.
                    tile_grid_row.append(' ')
                elif plan_char == '#':
                    # Wall tile.
                    tile_grid_row.append('#')
                elif plan_char == '@':
                    # Boulder actor.
                    tile_grid_row.append(' ')
                    actors.append(Boulder(tile_location))
                elif plan_char == '%':
                    # Crate actor.
                    tile_grid_row.append(' ')
                    actors.append(Crate(tile_location))
                elif plan_char == 'S':
                    # Starting spawn point of the avatar.
                    tile_grid_row.append(' ')
                    self.avatar = Avatar(tile_location)
                    actors.append(self.avatar)
                elif plan_char == 'E':
                    # The level exit.
                    tile_grid_row.append(' ')
                    actors.append(LevelExit(tile_location))
                elif plan_char == '*':
                    # The time warp.
                    tile_grid_row.append(' ')
                    actors.append(TimeWarp(tile_location))
                elif plan_char == '!':
                    # The gate.
                    tile_grid_row.append(' ')
                    actors.append(Gate(tile_location))
                else:
                    # Unknown plan tile symbol.
                    print "Unknown plan tile symbol '" + plan_char + "' at " + tile_location.__repr__()
                    assert(False)
                tile_location[0] += 1
            tile_grid.append(tile_grid_row)
            max_tile_grid_row_length = max(max_tile_grid_row_length, len(tile_grid_row))
            tile_location[1] += 1

        # Make the tile grid have all the rows the same length by filling them
        # with blank spaces (empty tiles).
        for tile_grid_row in tile_grid:
            while (len(tile_grid_row) < max_tile_grid_row_length):
                tile_grid_row.append(' ')

        self.tile_grid = tile_grid
        self.actors = actors
        for actor in self.actors:
            actor.level = self

    def update(self):
        next_level = False
        game_over = False
        restart_level = False

        if self.continuum_broke:
            restart_level = True

        elif self.travelling_time:
            self.play_time -= 1
            if self.play_time == 0:
                restart_level = True
            else:
                for actor in self.actors:
                    actor.update_time_travel()
        else:
            for actor in self.actors:
                actor.update()
            self.play_time += 1

            colliders = self.gather_actors_colliding(self.avatar)
            if not self.avatar.is_moving and len(colliders) == 1 and isinstance(colliders[0], TimeWarp):
                # The avatar has entered the time warp.
                self.travelling_time = True
                self.avatar = Avatar(self.avatar.grid_position)
                self.actors.append(self.avatar)
                self.avatar.level = self
                self.avatar.creation_time = self.play_time
                self.avatar.render()
                self.layer.flash()
            if not self.avatar.is_moving and len(colliders) == 1 and isinstance(colliders[0], LevelExit):
                # The avatar has entered the exit booth.
                next_level = True

        #if self.play_time > 60:
        #    restart_level = True

        if next_level:
            return GAME_NEXT_LEVEL
        elif game_over:
            return GAME_OVER
        elif restart_level:
            return GAME_RESTART_LEVEL
        else:
            return GAME_CONTINUE

    def render(self):
        print "rendering"
        self.render_tile_grid()
        for actor in self.actors:
            actor.render()

    def render_tile_grid(self):
        tile_location = [0, 0]
        for tile_grid_row in self.tile_grid:
            tile_location[0] = 0
            for tile in tile_grid_row:
                # Render the tile at location tile_location.
                if tile == "#":
                    #tiles = ["wall1a", "rocks1a", "rocks1b", "rocks2a", "rocks2b", "rocks2c", "water1a", "water1b", "water1c"]
                    #tile_pic = tiles[(tile_location[0]*tile_location[1]) % 9] + ".png",
                    #tile_pic = tiles[random.randint (0, 8)] + ".png"
                    tile_pic = "wall1a.png"
                    s = Sprite(tile_pic, position=transform_grid_to_screen_position(tile_location))
                    self.layer.add(s)
                tile_location[0] += 1
            tile_location[1] += 1

    def move_avatar(self, direction):
        """Try to begin the avatar movement in the specified direction.

        direction       0 - Up.
                        1 - Right.
                        2 - Down.
                        3 - Left.
        """
        self.avatar.move_avatar(direction)

    def gather_actors_in_tile(self, tile_position):
        """Return all the actors which collide with the specified tile."""
        gathered_actors = []
        for actor in self.actors:
            if actor.is_moving:
                actor_x1 = int(math.floor(actor.exact_position[0]))
                actor_x2 = int(math.ceil(actor.exact_position[0]))
                actor_y1 = int(math.floor(actor.exact_position[1]))
                actor_y2 = int(math.ceil(actor.exact_position[1]))
            else:
                actor_x1 = actor_x2 = actor.grid_position[0]
                actor_y1 = actor_y2 = actor.grid_position[1]
            if tile_position[0] >= actor_x1 and tile_position[0] <= actor_x2 and \
               tile_position[1] >= actor_y1 and tile_position[1] <= actor_y2:
                gathered_actors.append(actor)
        return gathered_actors

    def gather_actors_colliding(self, the_actor):
        """Return all the actors which collide with the specified actor."""
        if the_actor.is_moving:
            #the_actor_x1 = int(math.floor(the_actor.exact_position[0]))
            #the_actor_x2 = int(math.ceil(the_actor.exact_position[0]))
            #the_actor_y1 = int(math.floor(the_actor.exact_position[1]))
            #the_actor_y2 = int(math.ceil(the_actor.exact_position[1]))
            return []
        else:
            the_actor_x1 = the_actor_x2 = the_actor.grid_position[0]
            the_actor_y1 = the_actor_y2 = the_actor.grid_position[1]

        gathered_actors = []
        for actor in self.actors:
            if actor == the_actor:
                continue
            if actor.is_moving:
                #actor_x1 = int(math.floor(actor.exact_position[0]))
                #actor_x2 = int(math.ceil(actor.exact_position[0]))
                #actor_y1 = int(math.floor(actor.exact_position[1]))
                #actor_y2 = int(math.ceil(actor.exact_position[1]))
                return []
            else:
                actor_x1 = actor_x2 = actor.grid_position[0]
                actor_y1 = actor_y2 = actor.grid_position[1]
            if the_actor_x2 >= actor_x1 and the_actor_x1 <= actor_x2 and \
               the_actor_y2 >= actor_y1 and the_actor_y1 <= actor_y2:
                gathered_actors.append(actor)
        return gathered_actors

# =============================================================================
#
# Sample level
# Level should have grid size 26x18 tiles
#
# =============================================================================

########################## 26x18
level1_map_plan = """
##################
#                #
#    *     E     #
#                #
#   @   S  %     #
#                #
#   @            #
#                #
#                #
##################
"""

class Level1(LevelEngine):
    def __init__(self):
        super(Level1, self).__init__()
        self.map_name = '"People assume that time is strict progression of cause to effect, but *actually* from a non-linear, non-subjective viewpoint - it\'s more like a big ball of WIBBLY WOBBLY... time-y wimey... stuff."\n-- Professor When'
        self.background_image = "space2b.png"
        self.set_map_plan(level1_map_plan)
        self.next_level = 2

level2_map_plan = """
 ######
 # %  #
 #@@  #  ###
##@##@#  #E###
#  S# # ##  *#
## ## ### @###
 ###*      #
   ## ######
    ###
"""
########################## 26x18

class Level2(LevelEngine):
    def __init__(self):
        super(Level2, self).__init__()
        self.map_name = 'Perfection is the key...'
        self.background_image = "pavement1a.png"
        self.set_map_plan(level2_map_plan)
        self.next_level = 3

########################## 26x18
level3_map_plan = """
  ####   ##               
 # %   ##              #
 # @      #    ###   # #  
 #@##@#  #@##     ###  #  
#  S# # ##  *#####     #
 #  #  #  @##        ## 
  ##*      #      ### #   
#   # ##  #      #     ## 
 #  ##  #  #####     ##   
#     ##            #     
# ##      ## #######    # 
#   ###### # #          # 
##       #    ##########  
  #      # #            # 
   # #### #   ###  ##   #
   #       ###   ##  ###  
    ######     #       E# 
          ##### ########  
"""
########################## 26x18

class Level3(LevelEngine):
    def __init__(self):
        super(Level3, self).__init__()
        self.map_name = "Make your way home!"
        self.background_image = "map2.png"
        self.set_map_plan(level3_map_plan)


########################## 26x18
level4_map_plan = """
  ####   ## #       # #
 #     ##   # @#      S#
 #    #  ##    ###   # #
 # #  #    ##  @  ###  #
#  *# # ##   # # #     # #
 #  #  #   ##  #     ##  #
  #   @    # ##   ### #
#   # ##  #   #  #     ##
 #  ##  #  ## ##     ##
#    @##       #    #
# ##      ## ### ##     #
#   ###### # #    #     #
##   #   #   @## #*#####
  @      # #       #    ##
   # #### #   ###  #    #
 # #       ###   ##  ###
%E  ######     #       E#
  #       ##### ########
"""
########################## 26x18

class Level4(LevelEngine):
    def __init__(self):
        super(Level4, self).__init__()
        self.map_name = "Ene due rabe!"
        self.background_image = "map1.png"
        self.set_map_plan(level4_map_plan)

########################## 26x18
level5_map_plan = """
     #
     ######              
  #  #    #              
##@### ## #              
#     % # #   ###         
## ### ## ####@@@###     
# @@@@ ## #E@@@  @ #     
# S   %#@ # % @@@% #     
######   #####@ %  #     
#   @ #     # %    #     
# # %% @@ % #@@#####     
# ##    @   %  #         
# ########### ##         
#             #
###############
"""
########################## 26x18

class Level5(LevelEngine):
    def __init__(self):
        super(Level5, self).__init__()
        self.map_name = "And what if you can't find a time warp?"
        self.background_image = "level_5.png"
        self.set_map_plan(level5_map_plan)

levels = [Level1, Level2, Level3, Level4, Level5]

# =============================================================================
#
# Sample invocation
#
# =============================================================================

if __name__ == "__main__":
    level_1 = Level1()
    level_1.render()
